from typing import Tuple
from easypage.conn import Conn


class Storage:
    """
        Storage

        https://chromedevtools.github.io/devtools-protocol/tot/Storage
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def get_cookies(
            self,
            context_id: str,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取 同一个 Context 下的 cookies

        https://chromedevtools.github.io/devtools-protocol/tot/Storage/#method-getCookies

        :param context_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {'cookies': [{'name': '', 'value': '', 'domain': '', 'path': '', 'expires': -1, 'size': 161, 'httpOnly': True, 'secure': True, 'session': True, 'sameSite': 'Strict', 'priority': 'Medium', 'sameParty': False, 'sourceScheme': 'Secure', 'sourcePort': 443}]}
        """
        return self.__conn.send(
            method="Storage.getCookies",
            params={
                "browserContextId": context_id
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )



"""
    日志输出
"""
import logging
from easypage import settings


class Logger:
    def __init__(self, set_level: int = None):
        """

        :param set_level: 设置日志级别（如：logging.DEBUG）
        """
        if not set_level:
            set_level = logging.DEBUG

        # 创建logger对象
        logger = logging.getLogger('EasyPage')

        # 设置日志级别
        logger.setLevel(set_level)

        # 创建控制台处理器
        console_handler = logging.StreamHandler()

        # 设置控制台处理器的日志级别
        console_handler.setLevel(set_level)

        # 创建日志格式器
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # 将日志格式器添加到处理器
        console_handler.setFormatter(formatter)

        # 将处理器添加到logger
        logger.addHandler(console_handler)

        self.logger = logger

    def debug(self, msg: str, *args):
        if args:
            self.logger.debug(msg.format(*args))
        else:
            self.logger.debug(msg)

    def info(self, msg: str, *args):
        if args:
            self.logger.info(msg.format(*args))
        else:
            self.logger.info(msg)

    def warning(self, msg: str, *args):
        if args:
            self.logger.warning(msg.format(*args))
        else:
            self.logger.warning(msg)

    def error(self, msg: str, *args):
        if args:
            self.logger.error(msg.format(*args))
        else:
            self.logger.error(msg)

    def exception(self, msg: str, *args):
        if args:
            self.logger.exception(msg.format(*args))
        else:
            self.logger.exception(msg)

    def critical(self, msg: str, *args):
        if args:
            self.logger.critical(msg.format(*args))
        else:
            self.logger.critical(msg)


# if settings.SYSTEM_TYPE == 'linux':
#     logger = Logger()
# else:
#     from loguru import logger
logger = Logger()


from typing import Tuple
from easypage.conn import Conn


class Dom:
    """
        Dom

        https://chromedevtools.github.io/devtools-protocol/tot/Dom
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        启用 DOM 功能（查询元素、修改元素属性等）

        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-enable

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="DOM.enable",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_document(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取当前页面的完整 DOM 结构（DOM的根节点）

        学习地址：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getDocument

        注意：
        - 页面一但有变动就需要加载，否则获取的数据不准
        - 这里仅仅是获取了 backendNodeId

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {'nodeId': 1, 'backendNodeId': 1, 'nodeType': 9, 'nodeName': '#document', 'localName': '', 'nodeValue': '', 'childNodeCount': 1, 'children': [{'nodeId': 2, 'parentId': 1, 'backendNodeId': 2, 'nodeType': 1, 'nodeName': 'HTML', 'localName': 'html', 'nodeValue': '', 'childNodeCount': 2, 'children': [{'nodeId': 3, 'parentId': 2, 'backendNodeId': 3, 'nodeType': 1, 'nodeName': 'HEAD', 'localName': 'head', 'nodeValue': '', 'childNodeCount': 0, 'attributes': []}, {'nodeId': 4, 'parentId': 2, 'backendNodeId': 4, 'nodeType': 1, 'nodeName': 'BODY', 'localName': 'body', 'nodeValue': '', 'childNodeCount': 0, 'attributes': []}], 'attributes': [], 'frameId': '9958473737B20C99BBC0BAE47BA5F6E4'}], 'documentURL': 'about:blank', 'baseURL': 'about:blank', 'xmlVersion': '', 'compatibilityMode': 'QuirksMode'}
        """
        return self.__conn.send(
            method="DOM.getDocument",
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def query_selector(
            self,
            selector: str,
            node_id: int = None,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        选择元素

        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-querySelector

        :param selector: 选择器（支持 css、xpath）
        :param node_id: 节点 id（基于某个节点查找时需要）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        # 没有节点 id 的话，获取根节点的 id
        if not node_id:
            status, root = self.get_document()
            if not status or not root:
                return status, root

            node_id = root['root']['nodeId']

        return self.__conn.send(
            method="DOM.querySelector",
            params={
                "nodeId": node_id,
                "selector": selector
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def query_selector_all(
            self,
            selector: str,
            node_id: int = None,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        选择多个元素

        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-querySelectorAll

        :param selector: 选择器（支持 css、xpath）
        :param node_id: 节点 id（基于某个节点查找时需要）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        # 没有节点 id 的话，获取根节点的 id
        if not node_id:
            status, root = self.get_document()
            if not status or not root:
                return status, root

            node_id = root['root']['nodeId']

        return self.__conn.send(
            method="DOM.querySelectorAll",
            params={
                "nodeId": node_id,
                "selector": selector
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def resolve_node(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        将给定的 DOM 节点 ID 解析为对应的 JavaScript 对象

        学习地址：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-resolveNode

        建议传参：backendNodeId，不行的话 nodeId 也行

        可以获取到 objectId

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {"object":{"type":"object","subtype":"node","className":"HTMLDocument","description":"#document","objectId":"5606348279645422547.1.1"}}
        """
        return self.__conn.send(
            method="DOM.resolveNode",
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def describe_node(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        根据 nodeId 处理返回 dict 对象

        可以根据 nodeId、backendNodeId、objectId 获取

        学习地址：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-describeNode

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {'node': {'nodeId': 73, 'backendNodeId': 122, 'nodeType': 1, 'nodeName': 'IFRAME', 'localName': 'iframe', 'nodeValue': '', 'childNodeCount': 0, 'attributes': ['src', 'https://kunpeng-sc.csdnimg.cn/?timestamp=1645783940/#/preview/1012105?positionId=588&adBlockFlag=0&adId=1050061&queryWord=&spm=3001.5907&articleId=0', 'frameborder', '0', 'width', '100%', 'height', '400px', 'scrolling', 'no'], 'frameId': '216187998879F6BA0E5C47C69713FE99'}}
        """
        return self.__conn.send(
            method="DOM.describeNode",
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_attributes(
            self,
            node_id: int,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获取属性

        学习地址：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getAttributes

        需要通过以下方法转换成需要的字典
            attrs_list = attributes["attributes"]
            attrs = dict(zip(attrs_list[::2], attrs_list[1::2]))

        :param node_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="DOM.getAttributes",
            params={
                "nodeId": node_id
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def focus(
            self,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        聚焦

        可以根据 nodeId、backendNodeId、objectId 获取

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-focus

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method='DOM.focus',
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def set_attribute_value(
            self,
            name: str,
            value: str,
            node_id: int,
            raise_err: bool = False,
            need_callback: bool = False
    ) -> Tuple[bool, dict]:
        """
        修改属性

        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-setAttributeValue

        :param name: 属性名称
        :param value: 属性值
        :param node_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method='DOM.setAttributeValue',
            params={
                "nodeId": node_id,
                "name": name,
                "value": value,
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_box_model(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        返回给定节点的框
        返回的比如 content 有 4 个坐标，分别是
        - 左上顶点
        - 右上顶点
        - 右下顶点
        - 左下顶点

        可以根据 nodeId、backendNodeId、objectId 获取

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getBoxModel

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method='DOM.getBoxModel',
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_frame_owner(
            self,
            frame_id: str,
            raise_err: bool = False,
            need_callback: bool = True
    ) -> Tuple[bool, dict]:
        """
        获取 iframe 的 backendNodeId、nodeId

        学习链接:https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getFrameOwner

        :param frame_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="DOM.getFrameOwner",
            params={
                "frameId": frame_id
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_outer_html(
            self,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        获取页面含标签的内容

        可以根据 nodeId、backendNodeId、objectId 获取

        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getOuterHTML

        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="DOM.getOuterHTML",
            params=kwargs,
            raise_err=raise_err,
            need_callback=need_callback,
        )



"""
    response 是设计的获取到 body 之后才会触发事件
    为了最大化提高效率，只在需要的时候获取
"""
import re
from typing import Callable, Dict

from easypage.cdp import CDP
from easypage.request import Request
from easypage.response import Response


class Listener:

    def __init__(self, context_id: str):
        """

        :param context_id: 上下文 id
        """
        self._http_listen_on = False
        self.__context_id = context_id
        self.__request_listeners: Dict[Callable, dict] = {}
        self.__response_listeners: Dict[Callable, dict] = {}

    @property
    def context_id(self):
        return self.__context_id

    """
        添加
    """

    def request(self, callback: Callable):
        """
        请求监听（匹配所有请求）

        :param callback:
        :return:
        """
        self.__request_listeners[callback] = {
            "func": callback
        }
        self._http_listen_on = True

    def request_on_url(self, url: str, callback: Callable, is_regex: bool = True):
        """
        请求监听

        :param url: 待匹配的 url
        :param is_regex: 是否正则匹配，不是那就用包含关系（in 匹配）
        :param callback:
        :return:
        """
        self.__request_listeners[callback] = {
            "url": url,
            "func": callback,
            "is_regex": is_regex
        }
        self._http_listen_on = True

    def response(self, callback: Callable):
        """
        响应监听（匹配所有响应）

        :param callback:
        :return:
        """
        self.__response_listeners[callback] = {
            "func": callback
        }
        self._http_listen_on = True

    def response_on_url(self, url: str, callback: Callable, is_regex: bool = True):
        """
        响应监听

        :param url: 待匹配的 url
        :param is_regex: 是否正则匹配，不是那就用包含关系（in 匹配）
        :param callback:
        :return:
        """
        self.__response_listeners[callback] = {
            "url": url,
            "func": callback,
            "is_regex": is_regex
        }
        self._http_listen_on = True

    def response_on_mime_type(self, mime_type: str, callback: Callable, is_regex: bool = True):
        """
        响应监听（根据响应的 mime_type，如：text/html、image/png、application/json）

        :param mime_type: 待匹配的 mime_type
        :param is_regex: 是否正则匹配，不是那就用包含关系（in 匹配）
        :param callback:
        :return:
        """
        self.__response_listeners[callback] = {
            "mime_type": mime_type,
            "func": callback,
            "is_regex": is_regex
        }
        self._http_listen_on = True

    """
        触发
    """

    def _on_request(self, request: Request):
        """
        请求触发

        :param request:
        :return:
        """
        for func, config in self.__request_listeners.items():
            if "url" in config:
                is_regex = config["is_regex"]
                if is_regex:
                    if re.search(config["url"], request.url):
                        func(request)
                elif config["url"] in request.url:
                    func(request)
            elif "is_regex" not in config:
                func(request)

    def _on_response(self, response: Response, cdp: CDP):
        """
        响应触发

        :param response: 响应
        :param cdp: cdp
        :return:
        """
        run_functions = []

        for func, config in self.__response_listeners.items():
            if "url" in config:
                is_regex = config["is_regex"]
                if is_regex:
                    if re.search(config["url"], response.url):
                        run_functions.append(func)
                elif config["url"] in response.url:
                    run_functions.append(func)
            if "mime_type" in config:
                is_regex = config["is_regex"]
                if is_regex:
                    if re.search(config["mime_type"], response.mime_type):
                        run_functions.append(func)
                elif config["mime_type"] in response.mime_type:
                    run_functions.append(func)
            elif "is_regex" not in config:
                run_functions.append(func)

        # 判断要不要处理
        if len(run_functions) == 0:
            return

        # 先去触发获取响应体
        response._get_body()

        # 构造事件去触发响应
        def handle_response():
            """
            触发响应

            :return:
            """
            for rf in run_functions:
                rf(response)

        # 创建事件
        cdp.once(event=f"handle_response_{response.request_id}", callback=handle_response)

    """
        移除
    """

    def remove_request(self, callback: Callable):
        """
        移除请求监听

        :param callback:
        :return:
        """
        if callback in self.__request_listeners:
            del self.__request_listeners[callback]

    def remove_response(self, callback: Callable):
        """
        移除响应监听

        :param callback:
        :return:
        """
        if callback in self.__response_listeners:
            del self.__response_listeners[callback]

    """
        其它方法
    """

    def _event_trigger(self, event: str, enable: bool = False):
        """
        事件触发器

        :param event: 事件
        :param enable: 启用状态
        :return:
        """



"""
    使用了 EventEmitter 作为事件管理

    # 创建事件监听案例
    async def intercept(kwargs) -> None:
        print(kwargs)
        interception_id = kwargs["interceptionId"]

        await page_conn.send(
            method="Network.continueInterceptedRequest",
            params={"interceptionId": interception_id},
        )

    page_conn.on("Network.requestIntercepted", lambda event: loop.create_task(intercept(event)))

    异步这里代码结构有问题，仅针对 send 方法
        事件监听，监听到事件之后将会触发 loop.run_until_complete 操作
        同时同步程序也会有 loop.run_until_complete 就会造成报错：This event loop is already running
        需要对代码结构做处理，如有统一的事件发送地址，或者使用类似队列的结构
"""
import json
import typing
import asyncio
import websockets
from threading import Lock
from easypage import errors
from pyee import EventEmitter
from easypage.logger import logger
from websockets.legacy import client

TYPE_PAGE = "page"
TYPE_BROWSER = "browser"

DRIVER_TYPE = typing.Union[
    TYPE_BROWSER,
    TYPE_PAGE
]


class ConnAsync(EventEmitter):
    def __init__(self, address: str, driver_type: DRIVER_TYPE, driver_id: str, loop: asyncio.AbstractEventLoop = None):
        """
        # 页面连接
        ws://127.0.0.1:9223/devtools/page/87B827CF2CAF710C997699F858467545

        # 浏览器连接
        ws://127.0.0.1:9223/devtools/browser/0909caac-09eb-4f48-a481-5de5decb3583

        :param loop: 事件循环（没传的话，会自动创建）
        :param address: 连接地址，如：127.0.0.1:9223
        :param driver_type: 控制类型，如：TYPE_PAGE
        :param driver_id: 对应 id，如：87B827CF2CAF710C997699F858467545
        """
        super().__init__()
        self.__loop = loop or asyncio.get_event_loop()
        self.address = address
        self.driver_id = driver_id
        self.driver_type = driver_type
        self.driver_url = f"ws://{address}/devtools/{driver_type}/{driver_id}"

        self.__lock = Lock()
        self.__id = 0
        self.__events = {}  # 存储事件
        self.__callbacks = {}  # 管理所有回调
        self.__connected = False  # 是否已经连接上（client.WebSocketCommonProtocol 是否已创建）
        self.__ws = self.new_conn()  # ws 连接
        self.__conn: client.WebSocketCommonProtocol = None  # type: ignore # 实际操作的连接
        self.__recv_fut = self.__loop.create_task(self.__recv())  # 异步消息接收

    def new_conn(self) -> client.Connect:
        """
        创建新的 ws 连接，使用的时候需要创建 client.WebSocketCommonProtocol

        :return:
        """
        return client.connect(
            uri=self.driver_url,
            max_size=None,
            loop=self.__loop,
            ping_interval=None,
            ping_timeout=None
        )

    def send_async(self, method: str, raise_err: bool = False, params: dict = None) -> typing.Awaitable:
        """
        异步发送消息，返回一个需要 await 等待的对象

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param params:
        :return
        """
        # 发送消息
        msg = {
            "id": self.__next_id,
            "method": method,
            "params": params or dict()
        }
        self.__loop.create_task(self.__send(msg))

        # 构建 Future 对象返回
        callback = self.__loop.create_future()
        self.__callbacks[msg["id"]] = callback
        callback.raise_err: bool = raise_err  # type: ignore
        callback.error: Exception = errors.CDPRecvMsgError()  # type: ignore
        callback.method: str = method  # type: ignore
        return callback

    async def __send(self, msg: dict) -> None:
        """
        消息发送

        :param msg:
        :return:
        """
        # 等待连接建立
        while not self.__connected:
            await asyncio.sleep(0)

        # 发送消息
        msg_id = msg["id"]
        msg = json.dumps(msg, ensure_ascii=False)
        logger.debug("[SEND] {}", msg)

        try:
            await self.__conn.send(msg)
        except (websockets.ConnectionClosed, websockets.ConnectionClosedOK):
            logger.warning(f"连接已关闭：{self.driver_url}")

            # 获取回调对象并设置结果
            callback = self.__callbacks.get(msg_id, None)
            if callback and not callback.done():
                callback.set_result((False, dict()))
                await self.close()

    async def __recv(self) -> None:
        """
        接收消息

        :return:
        """
        async with self.__ws as conn:
            self.__conn = conn
            self.__connected = True
            while self.__connected:
                try:
                    msg = await self.__conn.recv()
                    if not msg:
                        break
                    await self.__onmessage(msg=msg)
                except (websockets.ConnectionClosed, ConnectionResetError):
                    logger.warning(f"连接已关闭：{self.driver_url}")
                    break
                await asyncio.sleep(0)

    async def __onmessage(self, msg: str) -> None:
        """
        接收到的消息处理

        :param msg:
        :return:
        """
        await asyncio.sleep(0)
        if "Not allowed" in msg:
            logger.warning(f"当前 driver 不支持该命令，请切换 driver，当前 driver 类型：{self.driver_type}")
        msg = json.loads(msg)

        # 消息回调
        if msg.get('id') in self.__callbacks:
            callback = self.__callbacks.pop(msg['id'])
            if msg.get('error'):
                logger.warning("[RECV] {}", msg)
                if callback.raise_err:
                    callback.error.args = (f"接收到错误消息：{msg['error']}",)  # 设置错误消息
                    callback.set_exception(callback.error)
                else:
                    callback.set_result((False, msg.get('error')))
            else:
                logger.debug("[RECV] {}", msg)
                callback.set_result((True, msg.get('result')))

        # 事件触发
        else:
            logger.debug("[RECV] {}", msg)
            method = msg.get('method', '')
            params = msg.get('params', {})
            self.emit(method, **params)
            print("接收到事件 -> ", method, self.__events.get(method))

    async def close(self) -> None:
        """
        关闭连接

        :return:
        """
        self.__connected = False
        await asyncio.sleep(0)

        # 未获取的结果设置异常
        for cb in self.__callbacks.values():
            if cb.raise_err:
                cb.error.args = (f'连接关闭：{cb.method}',)  # 设置错误消息
                cb.set_exception(cb.error)
            else:
                cb.set_result((False, f'连接关闭：{cb.method}'))

        self.__callbacks.clear()

        # 关闭连接
        if self.__conn:
            await self.__conn.close()
            self.__conn = None

        # 取消等待
        if not self.__recv_fut.done():
            self.__recv_fut.cancel()

    @property
    def __next_id(self) -> int:
        """
        获取消息 id

        :return:
        """
        self.__lock.acquire(timeout=2)
        self.__id += 1
        self.__lock.release()
        return self.__id



"""
    定义类型
"""
import sys
from typing import Optional

# 版本兼容
if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict


class Proxy(TypedDict, total=False):
    server: str
    bypass: Optional[str]  # Optional 表示可以为 None
    user: Optional[str]
    pwd: Optional[str]



"""
    所有的 CDP 命令

    https://chromedevtools.github.io/devtools-protocol/
"""
import typing

from easypage.conn import Conn
from easypage.cdp_method.dom import Dom
from easypage.cdp_method.page import Page
from easypage.cdp_method.input import Input
from easypage.cdp_method.fetch import Fetch
from easypage.cdp_method.target import Target
from easypage.cdp_method.network import NetWork
from easypage.cdp_method.browser import Browser
from easypage.cdp_method.runtime import Runtime
from easypage.cdp_method.storage import Storage
from easypage.cdp_method.security import Security
from easypage.cdp_method.emulation import Emulation
from easypage.cdp_method.css import CSS


class CDP:
    def __init__(self, conn: Conn):
        self.__conn = conn

    def on(self, event: str, callback: typing.Callable):
        """
        新增事件

        :param event:
        :param callback:
        :return:
        """
        self.__conn.on(event=event, f=callback)

    def once(self, event: str, callback: typing.Callable):
        """
        新增事件，并且只执行一次

        :param event:
        :param callback:
        :return:
        """
        self.__conn.once(event=event, f=callback)

    def emit(self, event: str, *args, **kwargs):
        """
        触发事件

        :param event: 事件名称
        :param args:
        :param kwargs:
        :return:
        """
        if event in self.__conn._events:
            self.__conn.emit(event=event, *args, **kwargs)

    def send(self, method: str, raise_err: bool = False, params: dict = None) -> typing.Tuple[bool, dict]:
        """
        同步发送消息

        使用：status, value = send(method="", raise_err=False, params={})

        返回的结果是有两个值：
            - status 表示当前消息是否有错误发生（raise_err 为 True 时直接抛出）
            - value 是对应浏览器返回的消息，是一个 dict

        :param method:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param params:
        :return
        """
        return self.__conn.send(
            method=method,
            raise_err=raise_err,
            params=params
        )

    def closed(self) -> bool:
        """
        返回连接是否关闭

        :return:
        """
        return self.__conn.closed()

    @property
    def target(self) -> Target:
        return Target(self.__conn)

    @property
    def page(self) -> Page:
        return Page(self.__conn)

    @property
    def browser(self) -> Browser:
        return Browser(self.__conn)

    @property
    def runtime(self) -> Runtime:
        return Runtime(self.__conn)

    @property
    def dom(self) -> Dom:
        return Dom(self.__conn)

    @property
    def input(self) -> Input:
        return Input(self.__conn)

    @property
    def emulation(self) -> Emulation:
        return Emulation(self.__conn)

    @property
    def storage(self) -> Storage:
        return Storage(self.__conn)

    @property
    def fetch(self) -> Fetch:
        return Fetch(self.__conn)

    @property
    def network(self) -> NetWork:
        return NetWork(self.__conn)

    @property
    def security(self) -> Security:
        return Security(self.__conn)

    @property
    def css(self) -> CSS:
        return CSS(self.__conn)



"""
    启动参数大全：https://peter.sh/experiments/chromium-command-line-switches/
    常用启动参数：https://blog.csdn.net/bigcarp/article/details/121142873

    与浏览器启动无关的可以做 settings 的默认值，否则不可以
"""
import re
import shutil
from pathlib import Path
from typing import List, Union
from easypage import settings, errors, definition


class _OptionsBasicOperations:
    def __init__(self):
        self._args = []
        self._args.extend(settings.BROWSER_START_ARGS)
        self._extensions = set()

    def add(self, args: Union[str, List[str]]) -> None:
        """
        加入启动参数

        :param args:
        :return:
        """
        if isinstance(args, str):
            args = [args]

        for a in args:
            a = a.strip()
            self.remove(a)
            self._args.append(a)

    def remove(self, args: Union[str, List[str]]) -> None:
        """
        移除启动参数（包含关系，只要含有关键词就删除）

        :param args:
        :return:
        """
        if isinstance(args, str):
            args = [args]

        self._args = [a1 for a1 in self._args if not any(a2.strip() in a1 for a2 in args)]

    def exists(self, arg: str) -> Union[str, bool]:
        """
        判断是否存在指定参数（包含关系）

        :param arg:
        :return: 存在返回，不存在返回 False
        """
        arg = arg.strip()

        for a in self._args:
            if arg in a:
                return a
        return False

    @property
    def args(self) -> list:
        """
        获取所有的配置

        :return:
        """
        return self._args

    def clear(self, default: bool = False) -> None:
        """
        清空所有配置

        :param default: 是否清空默认配置
        :return:
        """
        self._args.clear()

        if not default:
            self._args.extend(settings.BROWSER_START_ARGS)

    @property
    def extensions(self) -> set:
        """
        获取所有的插件

        :return:
        """
        return self._extensions


class BrowserOptions(_OptionsBasicOperations):
    def __init__(self):
        super().__init__()
        self._ip: str = None  # type:ignore # 设置浏览器启动ip
        self._port: int = None  # type:ignore # 设置浏览器启动端口
        self._user_data_dir: Path = None  # type:ignore # 设置用户数据目录
        self._delete_user_data_dir_when_stop: bool = True  # 设置用户数据目录是否删除
        self._browser_path: Path = None  # type:ignore # 设置浏览器路径
        self._proxy: definition.Proxy = None  # type:ignore # 设置代理
        self._user_agent: str = None  # type:ignore # 设置 user_agent
        self._headless: bool = False  # 设置浏览器是否隐身
        self._timeout: int = settings.TIMEOUT  # 设置全局超时时间

    """
        程序参数
    """

    def set_address(self, port: int, ip: str = "127.0.0.1"):
        """
        设置浏览器启动端口

        :param ip:
        :param port: 不设置的话，默认是 9876
        :return:
        """
        self._ip = ip.strip().lower().replace('localhost', '127.0.0.1')
        self._port = port

    def set_timeout(self, seconds: int):
        """
        设置全局的超时时间

        :param seconds:
        :return:
        """
        self._timeout = seconds

    def set_extension(self, path: Union[str, Path]):
        """
        添加插件

        :param path: 插件 crx 解压后的路径
        :return:
        """
        path = Path(path).absolute().resolve()

        if not path.exists():
            raise FileNotFoundError(path)

        self._extensions.add(str(path))

    def set_delete_user_data_dir_when_stop(self, enable: bool = True):
        """
        配置是否在结束时删除用户文件夹

        :param enable:
        :return:
        """
        self._delete_user_data_dir_when_stop = enable

    """
        浏览器参数
    """

    def set_user_data_dir(self, path: str, delete_if_exists: bool = False, delete_when_stop: bool = True) -> None:
        """
        设置用户数据目录
        注意：目录代表着配置信息，如果不删除可能配置不会生效！

        :param path:
        :param delete_if_exists: 启动时存在的话是否要删除
        :param delete_when_stop: 程序终止的时候是否删除
        :return:
        """
        path = Path(path).absolute().resolve()
        if delete_if_exists and path.exists():
            shutil.rmtree(path, ignore_errors=True)
        path.mkdir(parents=True, exist_ok=True)

        self.add(f"--user-data-dir={path}")
        self._user_data_dir = path
        self._delete_user_data_dir_when_stop = delete_when_stop

    def set_browser_path(self, path: str):
        """
        配置浏览器路径

        :param path:
        :return:
        """
        path = Path(path).absolute().resolve()

        if not path.exists():
            raise FileNotFoundError(path)
        elif not path.is_file():
            raise TypeError(f"浏览器路径不是文件：{path}")

        self._browser_path = path

    def set_user_agent(self, user_agent: str):
        """
        设置 user_agent

        :param user_agent:
        :return:
        """
        self._user_agent = user_agent.strip()
        self.add(args=f"--user-agent={user_agent.strip()}")

    def set_proxy_server(self, proxy: definition.Proxy):
        """
        设置代理

        这里使用了其它的方式设置含有账号密码的代理，而不是一开始设置
        浏览器启动时不支持 --proxy-server 设置账号密码代理，这一问题是 chromium 自身决定的，而不是开发问题

        :param proxy: 如：http://127.0.0.1:7890
        :return:
        """
        if re.search(r'.*?:.*?@.*?\..*', proxy["server"]):
            raise errors.ProxySetError(msg="""
        代理格式错误，如需要使用账号密码代理，请参照下面的两种方法：
            1、配置默认账号密码
            opts = BrowserOptions()
            opts.set_proxy_server(proxy = {"server": "","user": "","pwd": ""})
            
            2、使用 context 实现：
                browser.new_context(proxy = {"server": "","user": "","pwd": ""})
        """.format(proxy))

        if not proxy["server"].startswith("http"):
            proxy["server"] = "http://" + proxy["server"]

        self.add(args=f'--proxy-server={proxy["server"]}')
        self._proxy = proxy

    def set_window_size(self, width: int, height: int):
        """
        设置窗口大小

        :param width:
        :param height:
        :return:
        """
        self.add(args=f"--window-size={width},{height}")

    """
        涉及启用、禁用的
    """

    def set_headless(self, enable: bool = True):
        """
        设置浏览器是否隐身

        :param enable: 是否启用
        :return:
        """
        self._headless = enable
        enable and self.add(args="--headless=new")

    def set_no_sandbox(self, enable: bool = True):
        """
        设置 no_sandbox

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--no-sandbox")

    def set_incognito(self, enable: bool = True):
        """
        启用无痕模式

        新建上下文的话就等于无痕模式了

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--incognito")

    def set_ignore_certificate_errors(self, enable: bool = True):
        """
        忽略证书错误

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--ignore-certificate-errors")

    def set_disable_images(self, enable: bool = True):
        """
        禁用图片加载

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--blink-settings=imagesEnabled=false")

    def set_disable_js(self, enable: bool = True):
        """
        禁用 js 加载

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--disable-javascript")

    def set_disable_gpu(self, enable: bool = True):
        """
        禁用 gpu

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--disable-gpu")

    def set_disable_extensions(self, enable: bool = True):
        """
        禁用扩展

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--disable-extensions")

    def set_disable_software_rasterizer(self, enable: bool = True):
        """
        禁用软件光栅化器

        软件光栅化器是一种渲染技术，用于在没有硬件加速支持的系统上进行图形渲染
        禁用软件光栅化器意味着如果 GPU 不可用，应用程序将不会尝试使用 CPU 作为后备选项

        :param enable: 是否启用
        :return:
        """
        enable and self.add(args="--disable-software-rasterizer")

    def set_start_maximized(self, enable: bool = True):
        """
        设置浏览器启动时最大窗口

        TODO 测试和设置窗口大小会不会有冲突

        :param enable: 是否启用
        """
        enable and self.add(args="--start-maximized")

    """
        属性读取
    """

    @property
    def user_data_dir(self) -> Path:
        return self._user_data_dir

    @property
    def browser_path(self) -> Path:
        return self._browser_path

    @property
    def address(self) -> str:
        if self._ip is None or self._port is None:
            return ""
        return f"{self._ip}:{self._port}"

    @property
    def headless(self) -> bool:
        return self._headless

    @property
    def port(self) -> int:
        return self._port

    @property
    def ip(self) -> str:
        return self._ip

    @property
    def user_agent(self) -> str:
        return self._user_agent

    @property
    def timeout(self) -> int:
        return self._timeout

    @property
    def proxy(self) -> dict:
        if not self._proxy:
            return {}
        return self._proxy

    @property
    def delete_user_data_dir_when_stop(self) -> bool:
        return self._delete_user_data_dir_when_stop


if __name__ == '__main__':
    opts = BrowserOptions()
    opts.remove("disable")
    print(opts.args)



from typing import Tuple
from easypage.conn import Conn


class CSS:
    """
        CSS

        https://chromedevtools.github.io/devtools-protocol/tot/CSS
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            raise_err: bool = False,
            need_callback: bool = False
    ) -> Tuple[bool, dict]:
        """
        启用 css

        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-enable

        :param raise_err:
        :param need_callback:
        :return:
        """
        return self.__conn.send(
            method="CSS.enable",
            raise_err=raise_err,
            need_callback=need_callback
        )

    def disable(
            self,
            raise_err: bool = False,
            need_callback: bool = False
    ) -> Tuple[bool, dict]:
        """
        启用 css

        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-disable

        :param raise_err:
        :param need_callback:
        :return:
        """
        return self.__conn.send(
            method="CSS.disable",
            raise_err=raise_err,
            need_callback=need_callback
        )

    def get_computed_style_for_node(
            self,
            node_id: int,
            raise_err: bool = False,
            need_callback: bool = True
    ) -> Tuple[bool, dict]:
        """
        获取元素的 style

        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-getComputedStyleForNode

        :param node_id:
        :param raise_err:
        :param need_callback:
        :return:
        """
        return self.__conn.send(
            method="CSS.getComputedStyleForNode",
            params={
                "nodeId": node_id
            },
            raise_err=raise_err,
            need_callback=need_callback
        )



import asyncio
import functools
import typing
import traceback
from pathlib import Path
from typing import Union


def path_exists(path: str) -> Union[Path, bool]:
    """
    检查一个字符串是否是路径是否存在

    :param path:
    :return:
    """
    try:
        p = Path(path)
        if p.exists():
            return p
    except OSError:
        pass

    return False


def catch_error(f: typing.Callable):
    """
    错误补货装饰器，用于事件触发时将错误抛出，不然 pyee 模块事件执行错误没有报错

    :param f:
    :return:
    """

    def wrapper(*args, **kwargs):
        try:
            # 正常执行函数
            return f(*args, **kwargs)
        except:
            traceback.print_exc()

    return wrapper


# 将普通函数转为异步函数
def to_async(func):
    @functools.wraps(func)
    async def _wrapper(*args, **kwargs):
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(None, functools.partial(func, *args, **kwargs))
        return result

    return _wrapper


# 简单的双向链表
class DoublyLinkedDict:
    def __init__(self):
        self.__map1 = {}
        self.__map2 = {}

    def __setitem__(self, key, value):
        self.__map1[key] = value
        self.__map2[value] = key

    def __getitem__(self, key):
        if key in self.__map1:
            return self.__map1[key]
        elif key in self.__map2:
            return self.__map2[key]

        raise KeyError(f"不存在 {key}")

    def __delitem__(self, key):
        if key in self.__map1:
            key2 = self.__map1.pop(key)
            del self.__map2[key2]
        elif key in self.__map2:
            key2 = self.__map2.pop(key)
            del self.__map1[key2]

    def items(self):
        for k, v in self.__map1.items():
            yield k, v

    def keys(self):
        for k in self.__map1.keys():
            yield k

    def values(self):
        for v in self.__map1.values():
            yield v

    @property
    def datas(self) -> dict:
        return self.__map1



from typing import Tuple
from easypage.conn import Conn
from easypage.utils import other_utils


class Runtime:
    """
        Runtime

        https://chromedevtools.github.io/devtools-protocol/tot/Runtime
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def evaluate(
            self,
            js: str,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        执行 js 脚本

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-evaluate

        :param js: 路径或者 js 字符串
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {'id': 1, 'result': {'result': {'type': 'number', 'value': 2, 'description': '2'}}}
        """
        if other_utils.path_exists(js):
            with open(js, "r") as f:
                js = f.read()

        send_data = {
            'expression': js,
            "returnByValue": False,
            "awaitPromise": True,
            "userGesture": True
        }
        send_data.update(kwargs)

        return self.__conn.send(
            method="Runtime.evaluate",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def call_function_on(
            self,
            js: str,
            object_id: str,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        执行 js 脚本，可以指定 objectId 这样就相当于上下文是元素自身了

        学习链接：https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-callFunctionOn

        :param js: 路径或者 js 字符串
        :param object_id: 对应的对象（相当于上下文）
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        if other_utils.path_exists(js):
            with open(js, "r") as f:
                js = f.read()

        send_data = {
            'functionDeclaration': js,
            'objectId': object_id,
            "awaitPromise": True,
            "userGesture": True
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Runtime.callFunctionOn",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_properties(
            self,
            object_id: str,
            raise_err: bool = False,
            need_callback: bool = True,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        获取给定对象的属性

        https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-getProperties

        :param object_id:
        :param kwargs:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            "objectId": object_id,
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Runtime.getProperties",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )



"""
    默认公共配置，不做修改
"""
import sys
import logging
from typing import List

SYSTEM_TYPE: str = sys.platform.lower()  # 系统类型
BROWSER_START_IP: str = "127.0.0.1"  # 浏览器启动 ip
BROWSER_START_PORT: int = 9876  # 浏览器启动端口
BROWSER_START_ARGS: List[str] = [
    "--hide-crash-restore-bubble",  # 隐藏崩溃恢复提示泡泡
    "--no-default-browser-check",  # 禁用默认浏览器检查
    "--disable-features=PrivacySandboxSettings4",  # 禁用隐私沙箱设置的特定功能
    "--no-first-run",  # 禁用“首次运行”体验
    "--disable-infobars",  # 启动 Chrome 浏览器时禁用信息栏功能
    "--disable-suggestions-ui",  # 禁用建议功能的用户界面
    "--disable-background-timer-throttling",  # 禁用后台定时器的节流功能
    "--disable-breakpad",  # 禁用 Breakpad 异常报告系统
    "--disable-browser-side-navigation",  # 禁用浏览器端的页面导航，会向服务器发送完整的请求，然后加载新的页面
    "--disable-client-side-phishing-detection",  # 禁用钓鱼检测
    "--disable-default-apps",  # 禁用默认应用程序（不再自动使用系统中配置的默认应用程序，而是使用 Chrome 自带的相应功能或者行为）
    "--disable-dev-shm-usage",  # 禁用共享内存（容器化的场景下会对共享内存限制，可能导致无法正常工作，需要禁用）
    "--disable-features=site-per-process",  # 禁用站点隔离，减少资源消耗（所有的网站内容将会在同一个进程中运行）
    "--disable-hang-monitor",  # 禁用挂起监视器功能
    "--disable-popup-blocking",  # 禁用弹出窗口阻止功能
    "--disable-prompt-on-repost",  # 禁用在重新提交 POST 请求时出现的提示（如刷新的时候提交表单的情况）
    "--disable-sync",  # 禁用浏览器账户同步功能
    "--disable-translate",  # 禁用翻译功能
    "--metrics-recording-only",  # 启动 Chrome 浏览器时，浏览器将会在后台运行，但不会显示浏览器界面，也不会加载任何网页
    "--safebrowsing-disable-auto-update",  # 禁用浏览器自动更新
    "--use-mock-keychain",  # 用模拟的钥匙串（不会持久化保存，因此不会对系统的实际钥匙串产生任何影响）
    "about:blank",  # 显示一个空白页面
]  # 浏览器默认启动参数

TIMEOUT: int = 30  # 全局的超时时间（s）
LOGGER_LEVEL: int = logging.WARNING  # 日志输出级别
USER_AGENT: str = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36"  # 默认 ua
USER_DATA_DIR: str = f"./EasyPageTemp/user_data_"  # 默认用户文件夹（拼接的时候会加上 port）



"""
    浏览器 driver
"""
from typing import Tuple, Union, Dict, List, Callable
from easypage import definition, errors
from easypage.driver.page import PageDriver
from easypage.listener import Listener
from easypage.logger import logger
from easypage.conn import TYPE_BROWSER, TYPE_PAGE
from easypage.driver.driver import Driver
from easypage.options import BrowserOptions
from easypage.utils.other_utils import DoublyLinkedDict


class BrowserDriver(Driver):
    PAGES: Dict[str, "PageDriver"] = {}  # {page_id:page_driver}
    CONTEXTS: Dict[str, Union["BrowserDriver", "ContextDriver"]] = {}  # {context_id:context_driver}
    PAGE_CONTENT_MAPS: Dict[str, str] = {}  # {page_id:context_id}
    CONTENT_PAGE_MAPS: Dict[str, List[str]] = {}  # {context_id:[page_id]}

    # TODO CONTEXT 对 PAGE 是一对多的，不能用双向表

    def __init__(self, opts: BrowserOptions, driver_id: str):
        """

        :param opts:
        :param driver_id:
        """
        super().__init__(opts, TYPE_BROWSER, driver_id)

        # 获取基本信息
        _, targets = self.cdp.target.get_targets()
        self.context_id = targets['targetInfos'][0]['browserContextId']
        self._proxy = self._opts.proxy
        self.__class__.CONTEXTS[self.context_id] = self
        self.__create_page_driver(targets['targetInfos'])  # 启动时创建页面的 driver
        self.__listener = Listener(context_id=self.context_id)

        # 启用事件
        self.cdp.target.set_discover_targets()

        # 事件监听
        self.cdp.on(event="Target.targetCreated", callback=self.__on_target_target_created)
        self.cdp.on(event="Target.targetInfoChanged", callback=self.__on_target_target_info_changed)
        self.cdp.on(event="Target.targetDestroyed", callback=self.__on_target_target_destroyed)

    def __create_page_driver(self, target_infos: List[Dict]):
        """
        根据 target_infos 创建页面的连接

        :param target_infos: 启动时所有的页面信息
        :return:
        """
        for target_info in target_infos:
            if target_info.get("type") != "page":
                continue

            context_id = target_info["browserContextId"]
            proxy = self.__class__.CONTEXTS[context_id]._proxy if context_id in self.__class__.CONTEXTS else None

            # 创建 driver
            page_driver = PageDriver(
                opts=self._opts,
                page_id=target_info["targetId"],  # 页面的 targetId
                context_id=context_id,  # 页面的上下文 id
                context_driver=self.__class__.CONTEXTS[context_id],
                proxy=proxy,
            )
            self.__class__.PAGES[target_info["targetId"]] = page_driver
            self.__class__.PAGE_CONTENT_MAPS[target_info["targetId"]] = context_id

            if context_id not in self.__class__.CONTENT_PAGE_MAPS:
                self.__class__.CONTENT_PAGE_MAPS[context_id] = []
            self.__class__.CONTENT_PAGE_MAPS[context_id].append(target_info["targetId"])

    @property
    def listen(self) -> Listener:
        """
        事件监听

        :return:
        """
        return self.__listener

    def new_context(self, proxy: definition.Proxy = None, **kwargs) -> "ContextDriver":
        """
        创建新上下文（类似于隐身模式）

        :param proxy:
        :param kwargs:
        :return:
        """
        # 检查代理
        if proxy and not self._opts.proxy:
            raise errors.ProxySetError(
                "受限于 chromium 限制，上下文设置代理前一定要给全局设置代理，请使用 set_proxy 设置并重新启动浏览器！"
            )

        # 新建上下文
        _, context_data = self.cdp.target.create_browser_context(
            proxy=proxy.get("server") if proxy else None,
            proxy_pass_list=proxy.get("bypass") if proxy else None,
            **kwargs
        )

        # 存储上下文
        context_driver = ContextDriver(browser_driver=self, context_id=context_data["browserContextId"], proxy=proxy)
        self.__class__.CONTEXTS[context_data["browserContextId"]] = context_driver

        return context_driver

    def new_page(
            self,
            new_window: bool = False,
            context_id: str = None,
            **kwargs
    ) -> "PageDriver":
        """
        创建页面

        :param new_window: 是否打开新窗口
        :param context_id:
        :param kwargs:
        :return:
        """
        kwargs.update({
            "url": "about:blank",  # 必要参数
        })

        if new_window:
            kwargs["newWindow"] = True

        if context_id:
            kwargs["browserContextId"] = context_id

        _, target_data = self.cdp.target.create_target(
            **kwargs
        )

        return self.__class__.PAGES[target_data["targetId"]]

    def close(self):
        """
        关闭浏览器

        :return:
        """
        self.cdp.browser.close()

    """
       事件监听
    """

    def __on_target_target_created(self, *args, **kwargs):
        """
        Target.targetCreated 事件监听器

        :param args:
        :param kwargs:
        :return:
        """
        target_info = kwargs["targetInfo"]
        if target_info.get("type") != "page":
            return

        # 为页面创建 driver
        self.__create_page_driver(target_infos=[target_info])

    def __on_target_target_info_changed(self, *args, **kwargs):
        """
        Target.targetInfoChanged 事件监听器

        :param args:
        :param kwargs:
        :return:
        """

    def __on_target_target_destroyed(self, *args, **kwargs):
        """
        Target.targetDestroyed 事件监听器

        :param args:
        :param kwargs:
        :return:
        """
        target_id = kwargs["targetId"]

        if target_id in self.__class__.PAGES:
            if "target_id" in self.__class__.PAGES:
                del self.__class__.PAGES[target_id]

            if "target_id" in self.__class__.PAGE_CONTENT_MAPS:
                context_id = self.__class__.PAGE_CONTENT_MAPS.pop(target_id)
                self.__class__.CONTENT_PAGE_MAPS[context_id].remove(target_id)


class ContextDriver:

    def __init__(self, browser_driver: BrowserDriver, context_id: str, proxy: definition.Proxy = None):
        """

        :param browser_driver:
        :param context_id:
        :param proxy: 代理
        """
        self._browser_driver = browser_driver
        self.context_id = context_id
        self._proxy = proxy
        self.__listener = Listener(context_id=self.context_id)

    def new_page(
            self,
            new_window: bool = False,
            **kwargs
    ) -> "PageDriver":
        """
        创建页面

        :param new_window: 是否打开新窗口
        :param kwargs:
        :return:
        """
        return self._browser_driver.new_page(
            new_window=new_window,
            context_id=self.context_id,
            proxy=self._proxy,
            **kwargs
        )

    @property
    def listen(self) -> Listener:
        """
        事件监听

        :return:
        """
        return self.__listener

    def close(self):
        """
        关闭上下文

        :return:
        """
        self._browser_driver.cdp.target.close_target(target_id=self.context_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
    事件管理器

    用来管理 enable、disable

    TODO 可以试试装饰器模式
"""
from threading import Lock
from easypage.cdp import CDP
from easypage.logger import logger


class EventManager:
    def __init__(self):
        self.events = {}

    def enable(self, driver_id: str, cdp: CDP, event_name: str):
        """
        启用事件

        :param driver_id: 页面的 id
        :param cdp: cdp 对象
        :param event_name: 例如 css
        :return:
        """
        self.execute(driver_id, cdp, event_name, True, "enable")

    def disable(self, driver_id: str, cdp: CDP, event_name: str):
        """
        禁用事件

        :param driver_id: 页面的 id
        :param cdp: cdp 对象
        :param event_name: 例如 css
        :return:
        """
        self.execute(driver_id, cdp, event_name, False, "disable")

    def execute(self, driver_id: str, cdp: CDP, event_name: str, status: bool, func_name: str, **kwargs):
        """
        事件

        :param driver_id: 页面的 id
        :param cdp: cdp 对象
        :param event_name: 例如 css
        :param status: 启用还是禁用
        :param func_name: 启用的方法名称如：enable、disable
        :return:
        """
        # 判断有没有当前方法
        cdp_method = getattr(cdp, event_name, False)
        if cdp_method is False:
            logger.warning(f"cdp 没有事件 {event_name}")
            return

        func = getattr(cdp_method, func_name, False)
        if func is False:
            logger.warning(f"{event_name} 没有方法 {func_name}")
            return

        # 初始化数据
        if event_name not in self.events:
            self.events[driver_id] = {
                "lock": Lock(),
            }

        # 判断是否在启用状态
        with self.events[driver_id]["lock"]:
            if self.events[driver_id][event_name]:
                return

            # 设置启用状态并启用
            self.events[driver_id][event_name] = status
            func(**kwargs)



import atexit
import shutil
from subprocess import Popen
from typing import List, Dict
from easypage.logger import logger
from easypage.utils import browser_utils
from easypage.options import BrowserOptions
from easypage.driver.page import PageDriver
from easypage import settings, operate, definition
from easypage.driver.browser import BrowserDriver, ContextDriver


class EasyPage:
    def __init__(
            self,
            opts: BrowserOptions = None,
            log_level: int = None,
            reuse: bool = False,
            reuse_close: bool = False
    ):
        """

        :param opts: 配置
        :param log_level: 日志等级（如：logging 模块的 logging.WARNING）
        :param reuse: 是否复用已经存在的浏览器（根据 opts 内的端口判断复用，无法复用则创建新端口）
        :param reuse_close: 复用浏览器时，是否在结束时关闭浏览器并清理资源（默认不关闭）
        """
        # 配置日志等级
        log_level = log_level or settings.LOGGER_LEVEL
        # TODO: 临时处理，为了通过 loguru，后续需要修改
        # if not settings.SYSTEM_TYPE.startswith("win"):
        logger.logger.setLevel(log_level)

        self._opts = opts or BrowserOptions()
        self._reuse = reuse
        self._reuse_close = reuse_close

        self._proc: "Popen" = None  # type: ignore
        self._start_browser()

        self._browser_driver_id = browser_utils.parse_driver_id(operate.browser(opts.address))[0]
        self._browser_driver = BrowserDriver(opts=self._opts, driver_id=self._browser_driver_id)

    @property
    def browser(self) -> "BrowserDriver":
        return self._browser_driver

    @property
    def page(self) -> "PageDriver":
        """
        默认页面

        :return:
        """
        default_page_id = BrowserDriver.CONTENT_PAGE_MAPS[self.browser.context_id][0]
        return BrowserDriver.PAGES[default_page_id]

    @property
    def pages(self) -> List["PageDriver"]:
        """
        获取所有的当前页面

        TODO 需要修改成实时的，面对高并发网站可能会失败

        :return:
        """
        return list(BrowserDriver.PAGES.values())

    @property
    def contexts(self) -> List["ContextDriver"]:
        """
        获取指定上下文

        TODO 需要修改成实时的，面对高并发网站可能会失败

        :return:
        """
        return list(BrowserDriver.CONTEXTS.values())

    def page_infos(self) -> dict:
        """
        获取所有页面

        :return:
        """
        status, infos = self.browser.cdp.target.get_targets()
        if not status:
            return {}

        result_infos = {
            "page": []
        }
        for info in infos["targetInfos"]:
            if info["type"] == "page":
                result_infos["page"].append({
                    "targetId": info['targetId'],
                    "contextId": info['browserContextId'],
                })

        return result_infos

    def new_context(self, proxy: definition.Proxy = None, **kwargs) -> "ContextDriver":
        """
        创建新上下文（类似于隐身模式）

        :param proxy:
        :param kwargs:
        :return:
        """
        return self._browser_driver.new_context(proxy=proxy, **kwargs)

    def _start_browser(self):
        """
        启动或连接浏览器

        :return:
        """
        # 复用的话检查是否有可以复用的浏览器
        if not self._opts.address:
            self._opts.set_address(ip=settings.BROWSER_START_IP, port=settings.BROWSER_START_PORT)

        if self._reuse:
            status = browser_utils.port_in_use(port=self._opts.port, ip=self._opts.ip)
            if not status:
                logger.debug("浏览器复用失败，当前地址无已打开的浏览器：{}，将获取随机端口启动！", self._opts.address)
            else:
                logger.debug("浏览器复用成功：{}", self._opts.address)
                if self._reuse_close:
                    atexit.register(self.close)
                return

        # 打开新的浏览器
        from easypage import operate
        self._proc = operate.new_browser(opts=self._opts)
        if not self._reuse:
            atexit.register(self.close)

        if self._reuse and self._reuse_close:
            atexit.register(self.close)

        # 等待浏览器打开
        operate.wait_connection(opts=self._opts)

    def close(self):
        """
        关闭浏览器，清理资源，如删除资源文件夹

        :return:
        """
        # cdp 发送关闭浏览器
        try:
            self._browser_driver.close()
        except:
            pass

        # 杀掉进程
        try:
            if self._proc and self._proc.poll() is None:
                self._proc.terminate()
                self._proc.wait()
        except:
            pass

        # 清空资源文件夹
        if self._opts.delete_user_data_dir_when_stop and self._opts.user_data_dir:
            shutil.rmtree(self._opts.user_data_dir, ignore_errors=True)



from typing import Tuple, List
from easypage.conn import Conn


class Fetch:
    """
        Fetch

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch
    """

    def __init__(self, conn: Conn):
        self.__conn = conn

    def enable(
            self,
            handle_auth_requests: bool = False,
            patterns: List[str] = None,
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """

        允许发出 requestPaused 事件

        注意：这会导致请求暂停，需要调用 failRequest、fulfillRequest 或 continueRequest/continueWithAuth 之一

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-enable

        :param handle_auth_requests:
        :param patterns:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return: {"method":"Fetch.requestPaused","params":{"requestId":"interception-job-1.0","request":{"url":"https://www.baidu.com/","method":"GET","headers":{"Accept":"text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7","Upgrade-Insecure-Requests":"1","User-Agent":"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36","sec-ch-ua":"\"Chromium\";v=\"125\", \"Not.A/Brand\";v=\"24\"","sec-ch-ua-mobile":"?0","sec-ch-ua-platform":"\"Windows\""},"initialPriority":"VeryHigh","referrerPolicy":"strict-origin-when-cross-origin"},"frameId":"F096060E94EF02389AAE43C27A9F54A4","resourceType":"Document"}}
        """
        if patterns is None:
            patterns = [{"urlPattern": "*"}]

        return self.__conn.send(
            method="Fetch.enable",
            params={
                "handleAuthRequests": handle_auth_requests,
                "patterns": patterns
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def fail_request(
            self,
            request_id: str,
            error_reason: str,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        发出 failRequest 事件

        可能具有以下错误：
            Failed
            Aborted
            TimedOut
            AccessDenied
            ConnectionClosed
            ConnectionReset
            ConnectionRefused
            ConnectionAborted
            ConnectionFailed
            NameNotResolved
            InternetDisconnected
            AddressUnreachable
            BlockedByClient
            BlockedByResponse

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-failRequest

        :param request_id:
        :param error_reason: https://chromedevtools.github.io/devtools-protocol/tot/Network/#type-ErrorReason
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Fetch.failRequest",
            params={
                "requestId": request_id,
                "errorReason": error_reason
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def ful_fill_request(
            self,
            request_id: str,
            resp_code: int,
            resp_headers: dict,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        发出 fulfillRequest 事件（用来提供完整的响应内容）

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-fulfillRequest

        使用 Fetch.fulfillRequest 命令时，您可以提供完整的响应内容，包括状态码、响应头部和响应主体。
        这个命令用于完全替换原始请求的响应，您需要提供响应的所有部分。
        通常在您希望完全替换原始响应内容时使用。

        :param request_id:
        :param resp_code:
        :param resp_headers:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            'requestId': request_id,
            'responseCode': resp_code,
            'responseHeaders': resp_headers
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Fetch.fulfillRequest",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def continue_request(
            self,
            request_id: str,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        发出 continueRequest 事件（用来修改拦截的响应的部分内容）

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-continueRequest

        :param request_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            'requestId': request_id,
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Fetch.continueRequest",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def continue_with_auth(
            self,
            request_id: str,
            user: str,
            pwd: str,
            response: str = "ProvideCredentials",
            raise_err: bool = False,
            need_callback: bool = False,
    ) -> Tuple[bool, dict]:
        """
        发出 continueWithAuth 事件（提供身份验证）

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-continueWithAuth

        authChallengeResponse 参数详解：
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#type-AuthChallengeResponse

        :param request_id:
        :param user:
        :param pwd:
        :param response:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Fetch.continueWithAuth",
            params={
                "requestId": request_id,
                "authChallengeResponse": {
                    'response': response,  # Default、CancelAuth（取消验证）、ProvideCredentials
                    'username': user,
                    'password': pwd,
                },
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def continue_response(
            self,
            request_id: str,
            raise_err: bool = False,
            need_callback: bool = False,
            **kwargs
    ) -> Tuple[bool, dict]:
        """
        发出 continueResponse 事件（用来修改拦截的响应的部分内容）

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-continueResponse

        使用 Fetch.continueResponse 命令时，您可以修改拦截的响应的部分内容，而不是提供完整的响应。
        您可以使用这个命令来修改响应头部、响应状态码和响应主体等部分内容，而不需要提供完整的响应。
        这个命令用于对拦截的响应进行部分修改，而不需要重新构造整个响应。
        通常在您只需要修改响应的部分内容而不是完全替换时使用。

        :param request_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        send_data = {
            'requestId': request_id,
        }

        send_data.update(kwargs)

        return self.__conn.send(
            method="Fetch.continueResponse",
            params=send_data,
            raise_err=raise_err,
            need_callback=need_callback,
        )

    def get_response_body(
            self,
            request_id: str,
            raise_err: bool = False,
            need_callback: bool = True,
    ) -> Tuple[bool, dict]:
        """
        获得响应体

        https://chromedevtools.github.io/devtools-protocol/tot/Fetch/#method-getResponseBody

        :param request_id:
        :param raise_err: 接收到错误消息时，是否抛出，默认为 False
        :param need_callback: 接收到消息时，是否返回，默认为 True
        :return:
        """
        return self.__conn.send(
            method="Fetch.getResponseBody",
            params={
                "requestId": request_id
            },
            raise_err=raise_err,
            need_callback=need_callback,
        )



import json
from urllib import parse


class Request:
    def __init__(self, data: dict):
        self.request_id = data["requestId"]
        self.frame_id = data.get("frameId", "")
        self.resource_type = data["resourceType"] if "resourceType" in data else data["type"]
        self._request = data["request"]
        self._url_parse = parse.urlparse(self._request["url"])

    @property
    def url(self) -> str:
        return self._request["url"]

    @property
    def clean_url(self) -> str:
        return self._url_parse.scheme + "://" + self._url_parse.hostname + self._url_parse.path

    @property
    def method(self) -> str:
        return self._request.get("method", "")

    @property
    def headers(self) -> dict:
        return self._request.get("headers", {})

    @property
    def cookies(self) -> dict:
        cookie = self.headers.get("Cookie") or self.headers.get("cookie")
        if not cookie:
            return {}

        cookie_res = {}

        for c in cookie.split(";"):
            c = c.strip()
            index = c.index("=")
            cookie_res[c[:index]] = c[index + 1:]

        return cookie_res

    @property
    def post_data(self) -> dict:
        data = {}

        if self._request.get("hasPostData"):
            data = self._request.get("postData")
            if data:
                data = json.loads(data)

        return data

    @property
    def params(self) -> dict:
        data = {}

        if self._url_parse.query:
            for i in self._url_parse.query.split("&"):
                index = i.index("=")
                data[i[:index]] = i[index + 1:]

        return data

    def __repr__(self):
        return f"<Request method=\"{self.method}\" url=\"{self.url}\">"
